#ifndef OPluginH
#define OPluginH

#include "OUtils.h"
#include "OAdapter.h"

#include <windows.h>
#include <Forms.hpp>

#define CM_INIT_COMPLETE WM_USER + 1
#define CM_PLUGIN_CLOSE  WM_USER + 2

namespace OPlugin
{
    using namespace OAdapter;
    using namespace OUtils;

    template <class AdapterClass> class PACKAGE CPluginTmp: public TForm
    {
    protected:
    private:
        AdapterClass* const Adapter;
        virtual void __fastcall WndProc(Messages::TMessage &Message)
        {
            switch(Message.Msg)
            {
                case WM_SYSCOMMAND:
                    TForm::WndProc(Message);
                    if (Message.WParam == SC_CLOSE)
                    {
                        PostMessage(Callback->MainForm()->Handle,
                                    CM_PLUGIN_CLOSE,
                                    reinterpret_cast<WPARAM>(Handle), 0);
                    }
                    break;
                case CM_INIT_COMPLETE:
                    OnInitComplete(Message);
                    break;
                default:
                    TForm::WndProc(Message);
                    break;
            }
        }
    protected:
        bool Initiated;
        virtual void __fastcall InitComplete() = 0;
        virtual void __fastcall OnInitComplete(Messages::TMessage &Message)
        {
            InitComplete();
        }
    public:
        typedef AdapterClass AdapterClass;
        __property AdapterClass* Callback = {read = Adapter};
        __fastcall CPluginTmp(AdapterClass* const Adapter, TComponent *Owner):
                   TForm(Owner),
                   Adapter(Adapter),
                   Initiated(false)
        {
        }
    };

    class PACKAGE CPlugin: public CPluginTmp<CMFAAdapter>
    {
    public:
        __fastcall CPlugin(AdapterClass* const Adapter);

    };

    template <class PluginClass> class PACKAGE CPluginHandleTmp
    {
    public:
        typedef PluginClass PluginClass;
        typedef PluginClass::AdapterClass AdapterClass;
    public:
        HINSTANCE         DllHandle;
        AnsiString        DllName;
        PluginClass* const Plugin;
        CPluginHandleTmp(HINSTANCE DllHandle,
                         AnsiString DllName,
                         PluginClass* const Plugin):
                         DllHandle(DllHandle), DllName(DllName), Plugin(Plugin)
        {
        }
    };

    class PACKAGE CPluginHandle: public CPluginHandleTmp<CPlugin>
    {
    public:
        CPluginHandle(HINSTANCE DllHandle,
                      AnsiString DllName,
                      PluginClass* const Plugin);
    };

    template <class PluginHandleClass> class PACKAGE CPluginTableTmp: private map<int, PluginHandleClass*>
    {
    private:
        inline int Size()
        {
            return size();
        }
    protected:
        typedef PluginHandleClass::PluginClass  PluginClass;
        typedef PluginHandleClass::AdapterClass AdapterClass;
        typedef PluginHandleClass               PluginHandleClass;
        typedef map<int, PluginHandleClass*>    Map;
        AnsiString ClassName;
    public:
        __property int Count = {read = Size};

        CPluginTableTmp()
        {
        };

        PluginHandleClass* const operator[](int FuncID) const
        {
            return Find(FuncID);
        }

        bool Add(int FuncID, HINSTANCE DllHandle, AnsiString DllName, PluginClass* const Plugin)
        {
            try
            {
                const_iterator PH = find(FuncID);
                if (PH == end())
                {
                    Map::operator[](FuncID) = new PluginHandleClass(DllHandle, DllName, Plugin);
                    return true;
                }
                return false;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  Add  " + AnsiString(ClassName), E);
                return false;
            } catch (...)
            {
                CatchEllipsisMessage("   Add  " + AnsiString(ClassName));
                return false;
            }
        }

        bool Remove(int FuncID)
        {
            try
            {
                const_iterator PH = find(FuncID);
                if (PH != end())
                {
                    if (PH->second) delete PH->second;
                    erase(FuncID);
                    return true;
                }
                return false;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  Remove  " + AnsiString(ClassName), E);
                return false;
            } catch (...)
            {
                CatchEllipsisMessage("   Remove  " + AnsiString(ClassName));
                return false;
            }
        }

        inline PluginHandleClass* const Find(int FuncID) const
        {
            try
            {
                const_iterator PH = find(FuncID);
                if (PH != end())
                {
                    return PH->second;
                }
                return NULL;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  Find  " + AnsiString(ClassName), E);
                return NULL;
            } catch (...)
            {
                CatchEllipsisMessage("   Find  " + AnsiString(ClassName));
                return NULL;
            }
        }

    };

    class PACKAGE CPluginTable: public CPluginTableTmp<CPluginHandle>
    {
    };

    class PACKAGE CPluginFunc: private map<HWND, int>
    {
    public:
        const int operator[](HWND) const;
        bool Add(HWND PluginHandle, int FuncID);
        bool Remove(HWND PluginHandle);
        inline const int Find(HWND PluginHandle) const;
    };

    template <class PluginTableClass>
    class PACKAGE CPluginsTmp : private PluginTableClass,
                                private CPluginFunc
    {
    protected:
        typedef PluginTableClass::PluginClass       PluginClass;
        typedef PluginTableClass::AdapterClass      AdapterClass;
        typedef PluginTableClass::PluginHandleClass PluginHandleClass;
        typedef PluginTableClass                    PluginTableClass;
        typedef void __fastcall (__closure *RegisterPluginFunctionPointer)(const int, PluginClass* const );
        typedef void __fastcall (__closure *UnregisterPluginFunctionPointer)(const int);
        AnsiString ClassName;
    private:
        const RegisterPluginFunctionPointer   PRegister;
        const UnregisterPluginFunctionPointer PUnregister;
        AdapterClass* const Adapter;
    public:
        CPluginsTmp(const RegisterPluginFunctionPointer   RFP,
                    const UnregisterPluginFunctionPointer UFP,
                    AdapterClass* const Adapter):
                    PRegister(RFP),
                    PUnregister(UFP),
                    Adapter(Adapter)
        {
        }

        PluginTableClass::Count;

        //////////////////////////////////////
        PluginClass* Register(const int FuncID,
                              const AnsiString DllName)
        {
            try
            {
                PluginHandleClass *PH = PluginTableClass::Find(FuncID);
                if (PH)
                {
                    SetForegroundWindow(PH->Plugin->Handle);
                    return PH->Plugin;
                }
                HINSTANCE h = LoadLibrary(DllName.c_str());
                PluginClass *Plugin = NULL;
                if (h)
                {
                    typedef PluginClass* __stdcall (*PDllExec)(AdapterClass* const Adapter);
                    PDllExec dllExec = (PDllExec) GetProcAddress(h, "Execute");
                    if (dllExec)
                    {
                        Plugin = dllExec(Adapter);
                    } else
                    {
                        return NULL;
                    }
                } else
                {
                    return NULL;
                }
                PluginTableClass::Add(FuncID, h, DllName, Plugin);
                CPluginFunc::Add(Plugin->Handle, FuncID);
                PRegister(FuncID, Plugin);
                return Plugin;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  Register  " + AnsiString(ClassName), E);
                return NULL;
            } catch (...)
            {
                CatchEllipsisMessage("   Register  " + AnsiString(ClassName));
                return NULL;
            }
        }

        ////////////////////////////////////////
        bool Unregister(const HWND PluginHandle)
        {
            try
            {
                const int FuncID = CPluginFunc::Find(PluginHandle);
                if (!FuncID) return false;
                PluginHandleClass* const PH = PluginTableClass::Find(FuncID);
                if (!PH) return false;
                PUnregister(FuncID);
                HINSTANCE h = PH->DllHandle;
                typedef void __stdcall (*PFree)(PluginClass*);
                PFree dllExec = (PFree)GetProcAddress(h, "Free");
                if (dllExec)
                {
                    dllExec(PH->Plugin);
                }
                CPluginFunc::Remove(PluginHandle);
                PluginTableClass::Remove(FuncID);
                FreeLibrary(h);
                return true;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  Unregister  " + AnsiString(ClassName), E);
                return false;
            } catch (...)
            {
                CatchEllipsisMessage("   Unregister  " + AnsiString(ClassName));
                return false;
            }
        }

        ////////////////////////////////////////////////
        PluginClass* operator [](const int FuncID) const
        {
            try
            {
                PluginHandleClass *PH = PluginTableClass::Find(FuncID);
                if (PH) return PH->Plugin;
                return NULL;
            } catch (const Exception &E)
            {
                CatchExceptionMessage("  []  " + AnsiString(ClassName), E);
                return NULL;
            } catch (...)
            {
                CatchEllipsisMessage("   []  " + AnsiString(ClassName));
                return NULL;
            }
        }

    };

    class PACKAGE CPlugins: public CPluginsTmp<CPluginTable>
    {
    public:
        CPlugins(const RegisterPluginFunctionPointer,
                 const UnregisterPluginFunctionPointer,
                 AdapterClass* const);
    };

    template <class AdapterClass> class PACKAGE CPluginDialogTmp: public TForm
    {
    protected:

        //AdapterClass* const Adapter;
    public:
        typedef AdapterClass AdapterClass;
        __fastcall CPluginDialogTmp(AdapterClass* const Adapter, TComponent *Owner):
                   TForm(Owner),
                   Adapter(Adapter)
        {
        }
        AdapterClass* const Adapter;
    };
}

#endif
